/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/
#pragma once
#ifndef CHANNEL_BRIDGE_H
#define CHANNEL_BRIDGE_H

// Basisklassen
#include "log/df_log.h"

// Klassen fuer Multimasterprotokolle.
#include "channel_1_2.h" // Version 1/2
#include "channel_4.h"   // Version 4

#include <map>
#include <thread>
#include <mutex>
#include <string>

/** Auflistung aller möglicher Dateitype, welche von der Konvertierungsfunktion in der DLL verarbeitet werden können.
*/
enum class FileType
{
    Undefined, // detect file type by extension
    AES_AEL,
    XML,
};

struct Bind_info
{
    int state;
    int typeDevice;
    uint32_t serial;
    std::string ip;

    Bind_info() :
        state(0),
        typeDevice(0),
        serial(0)
    {
    }
};

struct Binding_info
{
    int connection_id;
    Bind_info bi;

    Binding_info() :
        connection_id( 0 )
    {
    }
};

class ChannelBridge : public CDFLog
{
public:
    // Konstruktor fuer Kanalinitialisierung
    ChannelBridge( int connectionNumber );
    virtual ~ChannelBridge();

    // Setzt die Kanalnummer
    inline void setConnectionNumber( int connectionNumber )
    {
        // Zugewiesene Kanalnummer
        m_connectionNumber = connectionNumber;
        m_logCheck.m_connectionNumber = connectionNumber;

        // Kanalnummer setzen
        m_channel4.setConnectionNumber( connectionNumber );
        // Kommunikationsbasis von _channel4 an andere uebergeben
        m_channel1_2.setConnection( m_channel4.connection() );

        // Laden der Initialisierungsdatei, ggf. Anpassungen wegen geänderter Verbindungsnummer.
        logIniRead();
        // Prüfen der Logdateinamen, ggf. Anpassung dieser.
        checkCurrentFilePaths();
    }

    // Setzt Klassenzustand
    inline void setActiveModeState(Datafox::Communication::ActiveModeState as)
    {
        m_channel4.setActiveModeState( as );
    }
    // Prueft ob Klasse bereits "in Verwendung"
    inline Datafox::Communication::ActiveModeState activeModeState()
    {
        return m_channel4.activeModeState();
    }
    // Gibt direkt zurück ob der Kanal Aktiv ist
    inline bool isActive()
    {
        return (activeModeState() == Datafox::Communication::ams_active);
    }

    Bind_info m_bindInfo;
    // Setzt die Information ob die Kanalklasse an ein bestimmtes Geraet gebunden ist.
    inline void setBind(Bind_info bi)
    {
        m_bindInfo = bi;
    }
    // Prueft ob Klasse alle Informationen fuer eine Kanalbindung hat.
    inline int isBind()
    {
        return m_bindInfo.state;
    }

    // Verbindungsannahme von uebergebenem Listen-Socket
    int connect(SOCKET *listenSocket, int connectionNumber, uint8_t deviceNumber = ProtocolBase::defaultDeviceNumberStandAlone() )
    {
        return m_channel4.connect(listenSocket, connectionNumber, deviceNumber);
    }

    // Setzen der CallBack-Funktion fuer Datensatzbenachrichtigung
    void setCallBackNotifyRecord(int (*CBNotifyRecord)(CNotifyRecord&))
    {
        m_channel4.setCallBackNotifyRecordAvailable( CBNotifyRecord );
    }

    // Main - Funktion fuer Befehlsausfuehrung in Workerthread
protected:

    //#################################
    // CallRoutinen fuer die Abarbeitung anstehender Kommunikationsbefehle
    //	virtual int CallFunction(short module, short func, uint8_t *data, int &count);
    // NotifyRoutine fuer die Abarbeitung von Benachrichtigungen unter den Klassen
    //	int CallFunctionNotify(short module, short func, uint8_t *data, int &count)
    //	{ return _channel4.CallFunctionNotify(module, func, data, count); }

    // Für das Projekt Spar bei Freudenberg-Id.
    // Da die DLL mit JNA eingebunden wurde ist eine mehrfachansteuerung von mehreren Threads abzufangen.
    std::mutex m_mutex_api;

public:
    std::thread::id m_thread_id_open;

    bool tryEnterCriticalSection();
    void enterCriticalSection();
    void leaveCriticalSection();

public:
    inline void setLastErrorNumber(uint8_t deviceNumber, int errorNumber, int *error_number_as_argument = nullptr )
    {
        if ( deviceNumber == 255 )
        {
            for ( auto iter = m_map_last_error_number.begin(); iter != m_map_last_error_number.end(); ++iter )
            {
                m_map_last_error_number[iter->first] = errorNumber;
            }
        }
        else
        {
            m_map_last_error_number[deviceNumber] = errorNumber;
        }

        // Fehlerwert auch ggf. dem Funktionsargument zuweisen.
        if ( error_number_as_argument )
        {
            *error_number_as_argument = errorNumber;
        }
    }
    inline int getLastErrorNumber(uint8_t deviceNumber)
    {
        return m_map_last_error_number[deviceNumber];
    }
private:
    std::map<uint8_t, int> m_map_last_error_number;

public:
    /////////////////////////
    // Schnittstelle

    bool isOpen();
    int open(uint8_t deviceNumber, int iType, const std::string &szComString, uint32_t iComValue, uint16_t iComTimeOut);
    int openAEIII(int iType, const char *szComString, uint32_t iComValue, uint16_t iComTimeOut);
    void close(void);

    /// Ermitteln der uebergebenen Parameter an open.
    inline void getOpenParams( int &comm_type, std::string &comm_string, uint32_t &comm_value, uint16_t &comm_timeout )
    {
        m_channel4.connection().openParameters( (ConnectionBase::ConnectionType&)comm_type, comm_string, comm_value, comm_timeout );
    }

    // Geraeteadresse pruefen.
    inline uint8_t cast_device_address( int32_t device_address )
    {
        uint8_t l_device_address = 0;
        if ( device_address > 0 && device_address < 255 )
        {
            l_device_address = static_cast<uint8_t>( device_address );
        }

        return l_device_address;
    }

    // Busnummer des Kommunikationspartners
    uint8_t setDeviceNumberReceiver(uint8_t deviceNumber); 
    uint8_t deviceNumberReceiver();

    // Busnummer der DLL - Klassen
    uint8_t setDeviceNumberSender(uint8_t deviceNumber); 
    uint8_t deviceNumberSender();

    // Schnittstelle nach Fehler bereinigen
    void resetWithError(int err);

    // Speicher bereinigen
    void reset();

public:
    /////////////////////////
    // Kommunikationshelper

    int read(uint8_t *buffer, uint32_t receiveCount, uint32_t &receivedCount);
    int write(const uint8_t *buffer, uint32_t transmitCount, uint32_t &transmittedCount);

    void setCallBackFunction(int (*pFunction)(void));
    int getProtocolVersion2(uint8_t deviceNumber, uint8_t detailed, int *piError);
    // Gibt die ermittelte Protokollversion zurueck
    uint8_t getProtocolVersion(uint8_t deviceNumber);
    uint8_t setProtocolVersion(uint8_t deviceNumber, uint8_t version);
    uint8_t m_protocol_version[251];
    bool m_prefer_AEIII;     // Bevorzugt auf AEIII+ pruefen


    /////////////////////////
    // Zugriff auf HANDLE/SOCKET

    /// Ermitteln des Schnittstellenhandle.
    HANDLE getComPort();
    /// Setzen der Schnittstellenparameter.
    int setComPort(HANDLE hCom, int iType, uint32_t iComValue, uint16_t iTimeOut);

    int comReleaseHandle();
    int comUseHandle(HANDLE h);


    /////////////////////////
    // Helper

    /////////////////////////
    // Befehle

    // Allgeimein

    // Anzahl der letzten Ermittlungsversuche halten.
    int m_device_poll_retry;
    // Geraeteerreichbarkeit pruefen
    int checkDevice(uint8_t deviceNumber, int *piError, int iPollRetry);

    // PZE Master IV warmstart.
    int restart(uint8_t deviceNumber, int *piError);
    // Reset des Gerätes
    int reset(uint8_t deviceNumber, int mode, int *piError);

    // Senden einer Bestaetigungsmeldung an das AEIII+
    int sendMessage(uint8_t deviceNumber, uint8_t ucDelayTime,
        uint8_t ucAction,	uint8_t ucAudio, const std::string &text, int *piError);

    // Senden eines Infotextes fuer Geraetestatus.
    int sendInfotext(uint8_t deviceNumber, const std::string &text, int *piError);

    // RTC - Wert schreiben / lesen
    int setDateTime(uint8_t deviceNumber, const uint8_t *pucDateTime, int *piError);
    int getDateTime(uint8_t deviceNumber, uint8_t *pucDateTime, int *piError);

    // Globale-/Systemvariable schreiben / lesen
    int setGlobVar(uint8_t deviceNumber, const char *name, int iType, const char *val, int *piError);
    int getGlobVar(uint8_t deviceNumber, const char *name, int iType, char *val, int iValueLength, int *piError);

    // Relais setzen/ruecksetzen und zustand abfragen.
    int closeRelay(uint8_t deviceNumber, int iNum, int iTime, int *piError);
    int openRelay(uint8_t deviceNumber, int iNum, int *piError);
    int getRelayState(uint8_t deviceNumber, int iNum, int *piState, int *piTime, int *piError);

    // Seriennummer auslesen
    int getSerialNum(uint8_t deviceNumber, int *piSerialNum, int *piError);
    // Firmwareversion aus Dateiheader lesen
    int getVersionFirmwareFromFile(const std::string& filePath, std::string& version, int *piError);

    /////////////////////////
    // System Device Log
    // Datensatz lesen
    int systemReadRecord( uint8_t deviceNumber, uint8_t *buffer, uint16_t &size, int32_t &error );
    // Datensatz quittieren.
    int systemQuitRecord(uint8_t deviceNumber, int *error);
    // Datensatzzeiger ruecksetzen
    int systemRestoreRecords(uint8_t deviceNumber, int *piError);

    // Aktuellen Status des Flash abrufen
    int getFlashStatus(uint8_t deviceNumber, int *piStatus, int *piError);
public:
    /////////////////////////
    // Datensatz- / Listenbeschreibungen

    int getTableDescription(uint8_t deviceNumber, uint8_t ucList, int *piError);
protected:
    //int getTables4(uint8_t deviceNumber, char *typeName);
    int getListName(uint8_t deviceNumber, uint8_t ucList, std::string &name, int *piError);


public:
    /////////////////////////
    // Datensaetze

    // Ermitteln des ersten Datensatzes
    int collectData(uint8_t deviceNumber, int *piError);
    // Ermitteln des naechsten Datensatzes
    int getDatensatz( uint8_t *buffer, uint16_t &size, int32_t &error );
    // Pruefen ob min. ein Datensatz im AEIII+ vorhanden.
    int checkData(uint8_t deviceNumber, int *piError);
    // Loescht alle Datensaetzte im AEIII+
    int clearData(uint8_t deviceNumber, int *piError);
    // Datensatzzeiger ruecksetzen
    int restoreRecords(uint8_t deviceNumber, int *piError);
    // Datensatz lesen
    int readRecord(uint8_t deviceNumber, uint8_t *buffer, uint16_t &size, int32_t &error);
    // Datensatz quittieren.
    int quitRecord(uint8_t deviceNumber, std::string &param, int *error);
    // Anzahl der Datensaetze oder verwendetes Speichervolumen.
    int recordVolume(uint8_t deviceNumber,  int &number, int &memoryUsage, int &memorySize);
protected:
    // Datensatzfunktionen, Variable und Puffer fuer letzten gelesenen.
    bool m_recordCollected; // Nur für Funktionspaar CollectData/GetDatensatz.
    uint8_t m_recordData[IV_SYS_RECORDS_MAX];
    uint16_t m_recordSize;
    int m_recordError;


public:
    /////////////////////////
    // Listen

    // Laden der Importierten Listendaten
    int writeListData(uint8_t deviceNumber, int *piError);
protected:
    int writeListData4(uint8_t deviceNumber);
    //int DownloadLists(uint8_t deviceNumber, int *piError);

public:
    /////////////////////////
    // Zutrittslisten (Version 1 / Version 2)

    // Laden der Importierten Listendaten fuer Zutrittskontrolle
    int writeEntranceListData(uint8_t deviceNumber, int iListNum, int *piError);
    int writeEntrance2ListData(uint8_t deviceNumber, int iListNum, int *piError);
protected:
    int writeEntrance2ListData4(uint8_t deviceNumber, uint8_t listNum);
    int writeEntrance2List4(uint8_t deviceNumber, uint8_t listNum);


public:
    /////////////////////////
    // Timeboylisten

    // Laden der Importierten Listendaten fuer Timeboy
    int writeTimeboyListData(uint8_t deviceNumber, uint8_t id, uint8_t num, int reserve, int *piError);
protected:
    int writeTimeboyListData4(uint8_t deviceNumber, uint8_t id, uint8_t num, int reserve);
    int writeTimeboyList4(uint8_t deviceNumber, uint8_t id, uint8_t num, int reserve);


public:
    /////////////////////////
    // Setup

    int writeStudioFile(uint8_t deviceNumber, const std::string &filePath, int *piError);
    int readStudioFile(uint8_t deviceNumber, const std::string &filePath, int *piError);

    // Veraenderungen an der Setupdatei vornehmen
    int modifyStudioFile(const std::string &sourceFile, const std::string &destFile, const std::string &changeData, int flags, int *piError);

    // Conversion
    int convertFile( const std::string &srcFile, FileType srcType, const std::string &dstFile, FileType dstType, int flags, int *piError );

protected:
    int writeSetupData(int *piError);
    int writeSetupData4(uint8_t deviceNumber);
    int writeSetupData4_2(uint8_t deviceNumber);
    int readSetupData(const std::string &filePath, int *piError);
    int readSetupData4(uint8_t deviceNumber);
    int readSetupData4_2(uint8_t deviceNumber);

private:
    int fileErrorToErrorCode( const int nFileError ) const;

    // Conversion
    bool isConvertible( FileType srcType, FileType dstType ) const;
    FileType getFileTypeByExtension( const std::string &file ) const;

public:
    /////////////////////////
    // Firmware
    int writeFileDevice(uint8_t deviceNumber, std::string &filePath, int *piError);
protected:
    int writeDevice4(uint8_t deviceNumber);

public:
    /////////////////////////
    // Module
    int writeModule(uint8_t deviceNumber, int type, const std::string &filePath, int *piError);
    int writeTextData(uint8_t deviceNumber, const std::string &filePath);
    int writeTouchKeyData(uint8_t deviceNumber, const std::string &filePath);
    int writeDisplayColors(uint8_t deviceNumber, const std::string &filePath);
    int writeWlanConfigData(uint8_t deviceNumber, const std::string &filePath);
    int readWlanConfigData(uint8_t deviceNumber, const std::string &filePath);
    int writeHardwareConfigData(uint8_t deviceNumber, const std::string &filePath);
    int readHardwareConfigData(uint8_t deviceNumber, const std::string &filePath);
    int writeFileData(uint8_t deviceNumber, const std::string &filePath);
    int readFileData(uint8_t deviceNumber, const std::string &filePath);
    int read_data_block( uint8_t deviceNumber, const std::string &name, char *data, uint32_t *length, int flags );

public:
    /////////////////////////
    // Log / Trace

    /// Setzen des Logdateinamen.
    void setLogFileName(const std::string &szLogFileName);

    /// Log anschalten.
    inline void setLogOn()  
    {
        m_channel4.setTraceMode();
    }

    /// Log ausschalten.
    inline void setLogOff() 
    {
        m_channel4.setTraceMode( TraceBase::TraceMode::off );
    }

    void setTraceFileName( const std::string &filePath )
    {
        m_channel4.setTraceFilePath( filePath );
    }

    void changeTraceFilePath( const std::string& file_path );

    /////////////////////////
    // Passwort
    void setPassword(const char *password)
    {
        m_channel4.setPasswordCommunication(password);
    }
public:
    int writeMemToFile(uint8_t deviceNumber, int type, int address, int size, const std::string &filePath);
    int writeFileToMem(uint8_t deviceNumber, const std::string &filePath);

public:	
    /////////////////////////
    // Kanalklassen fuer Protokollverwendung.
    CChannel1_2 m_channel1_2;
    Channel4 m_channel4;

    inline int resultToError(int iRes, int *piError)
    {
        if ( iRes < 0 )
        {
            // Kommunikationsfehler
            iRes = -iRes;
            // Fehlercode umsetzen
            *piError = iRes;
            iRes = 0;
        }
        else if ( iRes > 0 ) 
        {
            // Wenn der Modulfehler -Funktion unbekannt- ist,
            // dann ist das Geraet zwar aktiv der Befehl
            // wird jedoch nicht unterstuetzt.
            if ( IV_RESULT_MODULE( iRes ) == 2 )
            {
                *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else if ( IV_RESULT_MODULE( iRes ) == 16 )
            {
                // Angesprochener Slot ist nicht vorhanden.
                *piError = IV_ERROR_SLOT_NONE;
            }
            else if ( IV_RESULT_MODULE( iRes ) == 17 )
            {
                // Kein Timeboy im Slot vorhanden.
                *piError = IV_ERROR_NO_TIMEBOY;
            }
            else if ( IV_RESULT_MODULE( iRes ) == 18 )
            {
                // Timeboytyp noch nicht erkannt.
                *piError = IV_ERROR_TIMEBOY_NOT_READY;
            }
            else if ( IV_RESULT_MODULE( iRes ) == 19 )
            {
                // Es wird versucht ein TimeboyI-III anzusprechen.
                *piError = IV_ERROR_TIMEBOY_I_II_III;
            }
            else if ( IV_RESULT_MODULE( iRes ) == 20 ) 
            {
                // Ausführung konnte wegen belegtem Kommunikationskanal nicht weitergereicht werden.
                *piError = IV_ERROR_DOCKING_BUSY;
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 28 ) 
            {
                // Passwort fuer Kommunikation wird benoetigt.
                *piError = IV_ERROR_NEED_PASSWORD;
            }
            else if ( IV_RESULT_FUNCTION( iRes ) == 29 )
            {
                // Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                *piError = IV_ERROR_PASSWORD_LOCK;
            }
            else
            {
                // Funktionsfehler
                *piError = IV_ERROR_PROCESSING;
            }

            // Fehler
            iRes = 0;
        }
        else
        {
            // OK
            *piError = IV_ERROR_SUCCESS;
            iRes = 1;
        }

        return iRes;
    }
public:
    /////////////////////////
    // Schnittstelle zu Channel1/Channel2
    int getModus(uint8_t deviceNumber, int *error);

    /////////////////////////
    // Schnittstelle zu Channel4
    void setCallBackFunctionWithParams(int (*pFunction)(uint32_t, uint32_t));
    void setCallBackBlockTransfer( int32_t (*pFunction)(int32_t, int32_t, uint32_t, uint32_t, uint32_t, uint32_t) );
    uint32_t crc32(uint8_t *data, uint32_t size, uint32_t crc = 0xFFFFFFFF, bool closure = true);

    // MODULE_HARDWARE

    // Einfache Anfrage ob das Geraet antwortet.
    int ping(uint8_t deviceNumber);

    // Typ lesen
    int setTypeDevice(uint8_t deviceNumber, uint8_t type);
    int getTypeDevice(uint8_t deviceNumber, uint8_t &type);
    int setTypeBoard(uint8_t deviceNumber, uint8_t type);
    int getTypeBoard(uint8_t deviceNumber, uint8_t &type);
    int setTypeTCPIP(uint8_t deviceNumber, uint8_t val);
    int getTypeTCPIP(uint8_t deviceNumber, uint8_t &val);
    int setTypeModemMC(uint8_t deviceNumber, uint8_t val);
    int getTypeModemMC(uint8_t deviceNumber, uint8_t &val);
    int setTypeDisplay(uint8_t deviceNumber, uint8_t type);
    int getTypeDisplay(uint8_t deviceNumber, uint8_t &type);
    int setTypeKeypad(uint8_t deviceNumber, uint8_t type);
    int getTypeKeypad(uint8_t deviceNumber, uint8_t &type);
    int setTypeTransponder(uint8_t deviceNumber, uint8_t type);
    int getTypeTransponder(uint8_t deviceNumber, uint8_t &type);
    int setTypeMMC(uint8_t deviceNumber, uint8_t type);
    int getTypeMMC(uint8_t deviceNumber, uint8_t &type);
    int setTypeRadio(uint8_t deviceNumber, uint8_t type);
    int getTypeRadio(uint8_t deviceNumber, uint8_t &type);
    int setTypeGPS(uint8_t deviceNumber, uint8_t val);
    int getTypeGPS(uint8_t deviceNumber, uint8_t &type);
    int setTypeDCF77(uint8_t deviceNumber, uint8_t type);
    int getTypeDCF77(uint8_t deviceNumber, uint8_t &type);
    int setTypeFingerprint(uint8_t deviceNumber, uint8_t type);
    int getTypeFingerprint(uint8_t deviceNumber, uint8_t &type);
    int setTypeExternalReader(uint8_t deviceNumber, uint8_t val);
    int getTypeExternalReader(uint8_t deviceNumber, uint8_t &val);
    int setTypeTouch(uint8_t deviceNumber, uint8_t val);
    int getTypeTouch(uint8_t deviceNumber, uint8_t &val);
    int setTypeSensor(uint8_t deviceNumber, uint8_t val);
    int getTypeSensor(uint8_t deviceNumber, uint8_t &val);

    // Komponente lesen
    int setComponentRS485_0(uint8_t deviceNumber, uint8_t val);
    int getComponentRS485_0(uint8_t deviceNumber, uint8_t &val);
    int setComponentRS485_1(uint8_t deviceNumber, uint8_t val);
    int getComponentRS485_1(uint8_t deviceNumber, uint8_t &val);
    int setComponentUSB(uint8_t deviceNumber, uint8_t val);
    int getComponentUSB(uint8_t deviceNumber, uint8_t &val);
    int setComponentInAnalog(uint8_t deviceNumber, uint32_t val);
    int getComponentInAnalog(uint8_t deviceNumber, uint32_t &val);
    int setComponentInDigital(uint8_t deviceNumber, uint32_t val);
    int getComponentInDigital(uint8_t deviceNumber, uint32_t &val);
    int setComponentOutDigital(uint8_t deviceNumber, uint32_t val);
    int getComponentOutDigital(uint8_t deviceNumber, uint32_t &val);
    int setComponentBattery(uint8_t deviceNumber, uint8_t val);
    int getComponentBattery(uint8_t deviceNumber, uint8_t &val);
    int setComponentCharge(uint8_t deviceNumber, uint8_t val);
    int getComponentCharge(uint8_t deviceNumber, uint8_t &val);

    // Option schreiben/lesen
    int setOptionIOModule(uint8_t deviceNumber, uint8_t val);
    int getOptionIOModule(uint8_t deviceNumber, uint8_t &val);
    int setOptionTBAdaption(uint8_t deviceNumber, uint8_t val);
    int getOptionTBAdaption(uint8_t deviceNumber, uint8_t &val);
    int setOptionComServer(uint8_t deviceNumber, uint8_t val);
    int getOptionComServer(uint8_t deviceNumber, uint8_t &val);
    int setOptionTPWrite(uint8_t deviceNumber, uint8_t val);
    int getOptionTPWrite(uint8_t deviceNumber, uint8_t &val);
    int setOptionHTTP(uint8_t deviceNumber, uint8_t val);
    int getOptionHTTP(uint8_t deviceNumber, uint8_t &val);
    int setOptionLanguage(uint8_t deviceNumber, uint8_t val);
    int getOptionLanguage(uint8_t deviceNumber, uint8_t &val);
    int setOptionBiokeyUpdate(uint8_t deviceNumber, uint8_t val);
    int getOptionBiokeyUpdate(uint8_t deviceNumber, uint8_t &val);
    int setSetupOptions(uint8_t deviceNumber, uint16_t mask, uint16_t options);
    int getSetupOptions(uint8_t deviceNumber, uint16_t &mask, uint16_t &options);
    int setOptionFirmware(uint8_t deviceNumber, int mask, int options);
    int getOptionFirmware(uint8_t deviceNumber, int &mask, int &options);
    int setPayableOptions(uint8_t deviceNumber, uint8_t mask, uint8_t options);
    int getPayableOptions(uint8_t deviceNumber, uint8_t &mask, uint8_t &options);



    // HWConfig lesen
    int setHWConfig(uint8_t deviceNumber, uint8_t val);
    int getHWConfig(uint8_t deviceNumber, uint8_t &val);

    // Gesamte Hardwareinfo lesen
    int getHardwareInfo(uint8_t deviceNumber, StructHardwareInfo &hwInfo);
    // Gesamte Hardwareinfo lesen
    int getHardwareInfo( uint8_t deviceNumber, HardwareInfo &hardwareInfo );
    // Gesamte Hardwareinfo in String umsetzen.
    int getInfos(StructHardwareInfo hwInfo, std::string &info);


    // MODULE_SYSTEM

    // Systemreset; Speicherbereiche loeschen
    //	int SystemReset(uint8_t deviceNumber, uint16_t modus);

    // Firmwareversion lesen
    int getVersionFirmware(uint8_t deviceNumber, std::string &version);
    int getVersionBootloader(uint8_t deviceNumber, std::string &version);

    // Seriennummer 
    int setSerialNumber(uint8_t deviceNumber, uint32_t serial);
    int getSerialNumber(uint8_t deviceNumber, uint32_t &serial);

    // Datum/Uhrzeit setzen/lesen
    //	int SetRTC(uint8_t deviceNumber, char *time);
    //	int GetRTC(uint8_t deviceNumber, char *time);

    // Systemvariable setzen/lesen
    //	int SetSystemVariable(uint8_t deviceNumber, char *name, char *val);
    //	int GetSystemVariable(uint8_t deviceNumber, char *name, char *val);

    // Relais setzen/lesen
    //	int SetRelay(uint8_t deviceNumber, char *name, uint8_t busNum, uint8_t relayNum, uint8_t status, uint16_t time);
    //	int GetRelay(uint8_t deviceNumber, char *name, uint8_t busNum, uint8_t relayNum, uint8_t &status, uint16_t &time);

    // Datensegmentuebertragung beginnen/enden
    int beginSegment(uint8_t deviceNumber, uint16_t art, uint32_t &address, uint32_t size, uint32_t crc);
    int endSegment(uint8_t deviceNumber, uint16_t art, uint32_t crc);

    // Flash setzen/lesen (die ersten 16kByte Bootloader koennen nur gelesen werden)
    int setFlash(uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count);
    int getFlash(uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count);
    int getFlashPage(uint8_t deviceNumber, uint32_t pageNum, uint32_t offs, uint8_t *buffer, uint32_t &count);

    // EEPROM setzen/lesen
    int setEEPROM(uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count);
    int getEEPROM(uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count);

    // RAM setzen/lesen
    int setRAM(uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count);
    int getRAM(uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count);

    // PROGRAM setzen/(lesen unterbunden)
    //	int SetProgram(uint8_t deviceNumber, uint32_t address, uint8_t *buf, uint16_t count);
    //	int GetProgram(uint8_t deviceNumber, uint32_t address, uint8_t *buf, uint16_t count);

    // MMC setzen/lesen 
    int setMMC(uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count);
    int getMMC(uint8_t deviceNumber, uint32_t address, uint8_t *buf, int count);

    // Systembios aufrufen/verlassen
    int enterBios(uint8_t deviceNumber, uint8_t protocolVersion);
    int exitBios(uint8_t deviceNumber);

    // Interface setzen/lesen (RS232, TCPIP...)
    int setInterface(uint8_t deviceNumber, uint8_t type);
    int getInterface(uint8_t deviceNumber, uint8_t &type);

    // Baud-Rate setzen/lesen
    int setBaud(uint8_t deviceNumber, uint32_t br);
    int getBaud(uint8_t deviceNumber, uint32_t &br);

    // DeviceID setzen/lesen
    int setDeviceIdBios(uint8_t deviceNumber, uint16_t id);
    int getDeviceIdBios(uint8_t deviceNumber, uint16_t &id);

    // Bootloader aktivieren
    int enterBootloader(uint8_t deviceNumber, uint8_t protocolVersion);

    // Monitor Mode vom XPort
    int enterXPortMonitor(uint8_t deviceNumber, char *data, uint16_t &count);
    int enterXPortMonitorExt(uint8_t deviceNumber, char *data, uint16_t &count);
    int commandXPort(uint8_t deviceNumber, char *buf, uint16_t &count);
    int exitXPortMonitor(uint8_t deviceNumber);

    // Oeffnen einer Tabelle
    int tableOpen(uint8_t deviceNumber, const std::string &name, uint32_t options, int &handle, std::string &reserved);
    // Schliessen der Tabelle
    int tableClose(uint8_t deviceNumber, int handle);

    // Setzen eines Filters auf die Tabelle
    int tableSetFilter(uint8_t deviceNumber, int handle, const std::string &filter);
    // Lesen des Filters auf die Tabelle
    int tableGetFilter(uint8_t deviceNumber, int handle, std::string &filter);
    // Loeschen des Filters der Tabelle
    int tableRemoveFilter(uint8_t deviceNumber, int handle);

    // Abfragen der Datensatzanzahl
    int tableGetRowCount(uint8_t deviceNumber, int handle, uint32_t &currentCount, uint32_t &unsortedCount, uint32_t &deletedCount);
    // Abfragen der aktuellen Datensatznummer
    int tableGetCurrentRow(uint8_t deviceNumber, int handle, uint32_t &currentRow);
    // Ueberspringen von n Datensaetzen (offset kann negativ sein), ab top/aktuell/bottom
    int tableSetCurrentRow(uint8_t deviceNumber, int handle, int offset, uint16_t direction);

    // Aendern des aktuellen Datensatzes
    int tableSetCurrentRowData(uint8_t deviceNumber, int handle, uint16_t flags, char separator, const char *data, uint16_t dataSize);
    // Aendern eines Spaltenwerts, des aktuellen Datensatzes
    int tableSetCurrentColumnData(uint8_t deviceNumber, int handle, uint16_t flags, const char *name, const char *data, uint16_t dataSize);
    // Aendern eines Spaltenwerts bei allen aktuell selektieren Zeilen.
    int tableSetAllRowsToColumnData(uint8_t deviceNumber, int handle, uint16_t flags, const char *name, const char *data, uint16_t dataSize);

    // Abfragen des aktuellen Datensatzes
    int tableGetCurrentRowData(uint8_t deviceNumber, int handle, uint16_t flags, char separator, char *data, uint16_t &dataSize);
    // Abfragen eines Spaltenwerts, des aktuellen Datensatzes
    int tableGetCurrentColumnData(uint8_t deviceNumber, int handle, uint16_t flags, const char *name, char *data, uint16_t &dataSize);

    // Daten an die Listentabelle anfuegen.
    int tableAppendRowData(uint8_t deviceNumber, int handle, uint16_t flags, char separator, const char *data, uint16_t dataSize);
    // Loescht den aktuellen Datensatz
    int tableDeleteCurrentRow(uint8_t deviceNumber, int handle, uint16_t flags);
    // Loescht die aktuellen Datensaetze, welche durch den gesetzten Filter verfuegbar sind.
    int tableDeleteAvailableRows(uint8_t deviceNumber, int handle, uint16_t flags);

    // Abfragen einer Modulinformation
    int getInfo(uint8_t deviceNumber, const std::string &name, const std::string &params, std::string &info);

    // Zutrittsbuchung durchfuehren
    int entrance2Identification(uint8_t deviceNumber, const char *TM, const char *Id, const char *Pin, char *status, int &statusLength);
    // LED, Relais auf Tuermodul schalten
    int entrance2OnlineAction(uint8_t deviceNumber, const char *TM, uint32_t mask, uint8_t type, uint8_t duration);

    // Transparentmodus aktivieren/deaktivieren
    int openComServerMode(uint8_t deviceNumber, const char *comm, uint32_t baudRate, uint8_t byteSize, uint8_t parity, uint8_t stopBits, uint16_t flags, uint16_t timeoutToClose);
    int closeComServerMode(uint8_t deviceNumber);

    // Aktive Meldungen aktivieren/deaktieren
    int setNotifyOnDevice(uint8_t deviceNumber, int type, int state);

    // Softwareinformation der Anwendungsfirmware abfragen
    int getSoftwareInfo(uint8_t deviceNumber, uint8_t *swi, int &infoLength);

    // Passwort pruefen
    int passwordCheck(uint8_t deviceNumber, char *password);
    // Passwortschlüssel abrufen
    int getPasswordKey(uint8_t deviceNumber, char *key);

    // Virtuellen Tastendruck ausführen.
    int pressVirtualKey(uint8_t deviceNumber, uint8_t vk, uint8_t flags, int *piError);

    // MODULE_TABLE
#define BIOKEY_OK                                (0x80)  ///< letzter Befehl erfolgreich
#define BIOKEY_ERROR                             (0x81)  ///< unerwarteter Fehler aufgetreten
#define BIOKEY_BAD_IMAGE                         (0x82)  ///< kein oder schlechtes Bild
#define BIOKEY_BAD_FPT                           (0x83)  ///< schlechtes Template, nicht genug Merkmale
#define BIOKEY_DB_FULL                           (0x84)  ///< Datenbank voll
#define BIOKEY_DB_EMPTY                          (0x85)  ///< Datenbank leer
#define BIOKEY_IDS_USED                          (0x86)  ///< übergebene PIDs werden schon benutzt
#define BIOKEY_IDS_UNKNOWN                       (0x87)  ///< übergebene PIDs existieren nicht
#define BIOKEY_RECOGNIZED                        (0x88)  ///< Person wurde erkannt
#define BIOKEY_NOT_RECOGNIZED                    (0x89)  ///< Person wurde nicht erkannt
#define BIOKEY_NO_IMAGE                          (0x8A)  ///< es wurde kein Bild vom Sensor aufgenommen
#define BIOKEY_NO_MATCH                          (0x8B)  ///< kein Treffer
#define BIOKEY_AUTH_FAILED                       (0x8C)  ///< falsches Passwort
#define BIOKEY_FPT_EXISTS                        (0x8D)  ///< an attempt to enroll with message cmd 0x26 failed because a template that matches the newly acquired one was found in the database
#define BIOKEY_FPT_ENROLLED                      (0x8E)  ///< a newly acquired fingerprint has been enrolled successfully
#define BIOKEY_TIMEOUT                           (0x8F)  ///< a timeout occurred during an enroll/identify process
#define BIOKEY_FPT_READ                          (0x90)  ///< a fingerprint was read during “Enroll best of three” procedure
    // Datensatz an Fingerprint-Tabelle anfuegen.
    int fingerprintAppendRecord(uint8_t deviceNumber, uint8_t type, const uint8_t *fingertemplate, uint16_t size);
    // Datensatz aus Fingerprint-Tabelle lesen.
    int fingerprintGetRecord(uint8_t deviceNumber, uint8_t &type, uint32_t pid, uint8_t fid, uint8_t *fingertemplate, uint16_t &size);
    // Datensatz aus Fingerprint-Tabelle loeschen.
    int fingerprintDeleteRecord(uint8_t deviceAddressremote, uint32_t pid, uint8_t fid);
    // Tabelle mit Fingerprintdaten erzeugen
    int fingerprintList(uint8_t deviceNumber, int flags, uint32_t &recordCount);
    // Backup der Fingerprints erstellen
    int fingerprintBackup(uint8_t deviceNumber, const std::string &filePath, int flags);
    // Wiederherstellen der Fingerprints
    int fingerprintRestore(uint8_t deviceNumber, const std::string &filePath, int flags);

    // Systemflags setzen/lesen
    int setSystemFlags(uint8_t deviceNumber, uint32_t flags);
    int getSystemFlags(uint8_t deviceNumber, uint32_t &flags);

    // MODULE_APPLICATION
    // Setzt den Font für die Anzeige der Message und Info - Text
    int setTypeFont(uint8_t deviceNumber, int iType, int *piError);

    // Setzt das Passwort fuer die verschluesselte Kommunikation
    int setCommunicationPassword(const char *password, uint8_t size, uint32_t sessionTimeout, bool global);
    int isGlobalCommunicationPassword();

private:
    int m_protocol_version_record_description;
    int m_protocol_version_list_description;

public:
    // Funktionen für Pufferzugriff.
    int GetTableCount(uint8_t ucList = 0);

    int GetTableRecord(uint8_t ucNum, std::string& szName, 
        uint8_t *pucFieldCount, uint8_t *pucVariable = nullptr, 
        uint8_t ucList = 0);

    int GetTableField(uint8_t ucTableNum, uint8_t ucFieldNum, 
        std::string& szName, uint8_t *pucType, 
        uint8_t *pucLength, uint8_t ucList = 0);

    void ClearListBuffer();
    int MakeListe(int ListNum, int ListCnt, int ListSize, const uint8_t *Buf);

    void ClearEntranceListBuffer(int iListNum = -1);
    int MakeEntranceList(int ListNum, int ListCnt, int ListSize, const uint8_t *Buf);

    void ClearEntrance2ListBuffer(int iListNum = -1);
    int MakeEntrance2List(int ListNum, int ListCnt, int ListSize, const uint8_t *Buf);

    void ClearTimeboyListBuffer(uint8_t id = 0xFF, uint8_t num = 0xFF);
    int MakeTimeboyList(uint8_t id, uint8_t num, uint16_t recordCount, const uint8_t *buf, uint32_t size, int flags, int *error);

    bool WriteListToBinFile( uint8_t listNum, const char *filePath );
    bool WriteListToTextFile( uint8_t listNum, const char *filePath );
    bool WriteListEntrance2ToBinFile( uint8_t listNum, const char *filePath );
    bool WriteListEntrance2ToTextFile( uint8_t listNum, const char *filePath );
    bool WriteListTimeboyToBinFile( uint8_t listId, uint8_t listNum, const char *filePath );

    std::string GetFileDeviceFileName();
    std::string GetFileDeviceError();
    std::string GetFileDeviceInfo();
    FileDevice &GetFileDevice();

    SoftwarebuildVersion &minimalUpdateVersion();
    std::list<std::string> &blacklistVersions();

    // Zutrittsbuchung durchfuehren
    int accessControlIdentification(uint8_t deviceNumber, int masterId, const char *TM, const char *Id, const char *Pin, char *status, int &statusLength);
    // LED, Relais auf Tuermodul schalten
    int accessControlOnlineAction(uint8_t deviceNumber, int masterId, const char *TM, uint32_t mask, uint8_t type, uint8_t duration);
    int accessControlKnobCommand( uint8_t deviceNumber, uint8_t knobCommand, uint16_t masterId, uint8_t knobId, const char *params, uint8_t *data, int *sizeData );

    int customCommand( uint8_t deviceNumber, uint32_t idendification, uint32_t customCommand, const char *dataIn, int sizeIn, char *dataOut, int *sizeOut );

    int startTransparencyMode( uint8_t deviceNumber, uint16_t id, uint16_t flags, uint32_t sizeReadBuffer, uint32_t sizeWriteBuffer, int timeout, int *errorNumber );
    int stopTransparencyMode( uint8_t deviceNumber, int *errorNumber );
    int isTransparencyMode();

    int ping( uint8_t deviceNumber, int attempts, int *errorNumber );

public:
    ControllerSegment::BlockTransferState getBlockTransferState(uint8_t deviceNumber)
    {
        ControllerSegment &controllerSegment = m_channel4.controllerSegment(deviceNumber);
        return controllerSegment.transferState;
    }

    int getBlockTransferActiveCount(ControllerSegment::BlockTransferType blockType)
    {
        return m_channel4.controllerSegmentActiveCount(blockType);
    }

    ControllerSegment::BlockTransferType getBlockTransferType(uint8_t deviceNumber)
    {
        ControllerSegment &controllerSegment = m_channel4.controllerSegment(deviceNumber);
        return controllerSegment.transferType;
    }

    void setBlockTransferSettings(uint8_t deviceNumber, uint32_t typesAllowed, uint16_t duration)
    {
        ControllerSegment &controllerSegment = m_channel4.controllerSegment(deviceNumber);
        controllerSegment.allowedTransferTypes = typesAllowed;
        if (typesAllowed != ControllerSegment::btt_none)
        {
            controllerSegment.duration = std::chrono::seconds( duration );
        }
        else
        {
            controllerSegment.duration = std::chrono::seconds( 0 );
        }

        if ( deviceNumber != 255 )
        {
            // Ggf. Flag fuer die uebernahme vom Global-Objekt zuruecksetzen.
            controllerSegment.transferFlags.settingsFromGlobal = false;
        }
        else
        {
            // Wird Global-Objekt auf 0 gesetzt muessen alle die dieses Flag bekommen haben auch zuruecksetzen.
            m_channel4.resetControllerSegmentGlobalDuration();
        }
    }

    void getBlockTransferSettings(uint8_t deviceNumber, uint32_t *typesAllowed, uint16_t *duration)
    {
        ControllerSegment &controllerSegment = m_channel4.controllerSegment(deviceNumber);
        *typesAllowed = controllerSegment.allowedTransferTypes;
        *duration = static_cast<uint16_t>( controllerSegment.duration.count() );
    }

    int blockTransferResume(uint8_t deviceNumber)
    {
        int res = IV_ERROR_SUCCESS;
        ControllerSegment &controllerSegment = m_channel4.controllerSegment(deviceNumber);

        // Ist Transfer aktiv?
        if ( controllerSegment.transferState != ControllerSegment::BlockTransferState::active )
        {
            return -IV_ERROR_BLOCK_TRNASFER_INACTIVE;
        }

        switch ( controllerSegment.transferType )
        {
        case ControllerSegment::btt_none:
            res = -IV_ERROR_FAIL;
            break;

        case ControllerSegment::btt_list:
            {
                int errorNumber = IV_ERROR_SUCCESS;
                res = writeListData(deviceNumber, &errorNumber);
                if (res == 0)
                    res = -errorNumber; // Fehlerwert zurueckgeben.

                break;
            }

        case ControllerSegment::btt_access_control_list:
            {
                int errorNumber = IV_ERROR_SUCCESS;
                res = writeEntrance2ListData(deviceNumber, -1, &errorNumber);
                if (res == 0)
                    res = -errorNumber; // Fehlerwert zurueckgeben.

                break;
            }

        case ControllerSegment::btt_write_file:
            {
                res = writeFileData(deviceNumber, std::string());
                break;
            }

        case ControllerSegment::btt_read_file:
            {
                res = readFileData(deviceNumber, std::string());
                break;
            }

        case ControllerSegment::btt_write_setup_file:
            {
                res = m_channel4.writeSetupData(deviceNumber, std::string());
                break;
            }

        case ControllerSegment::btt_read_setup_file:
            {
                res = m_channel4.readSetupData(deviceNumber, std::string());
                break;
            }

        case ControllerSegment::btt_mask_known:
            res = -IV_ERROR_FAIL;
            break;
        }

        return res;
    }

    int blockTransferDiscard(uint8_t deviceNumber)
    {
        ControllerSegment &controllerSegment = m_channel4.controllerSegment(deviceNumber);

        // Ist Transfer aktiv?
        if ( controllerSegment.transferState != ControllerSegment::BlockTransferState::active )
        {
            return -IV_ERROR_BLOCK_TRNASFER_INACTIVE;
        }

        if ( isOpen() )
        {
            blockTransferResume( deviceNumber );
        }

        // Sicherheitshalber noch Inactiv setzen, falls Kommunikationsfehler oder Schnittstelle geschlossen.
        controllerSegment.transferState = ControllerSegment::BlockTransferState::inactive;

        return IV_ERROR_SUCCESS;
    }
};

#endif
